from .decode_head import decode_yolo_head
try:
    from .inference import InferenceSession
    from .device import KneronDevice
except:
    print("[warning] kp is not install")

try:
    from .convert import onnx_to_nef, combine_nef_files
except ImportError:
    print("[Warning] Kneron toolchain not found. Please install it when want to use the convert function.")